"""Module"""
from dataclasses import dataclass
from dataclasses_json import dataclass_json


@dataclass_json
@dataclass
class DaikinProfile:
    """Daikin user profile object"""

    name: str
    lastName: str
    toc: bool
    commercial: bool
    support: bool
    commercial_date: str
    support_date: str
    email: str



"""
Module

TODO: Figure out dependency import paths
using sys.path.append("/workspaces/core/charles_dev") isn't goning to work for prod


"""
import sys

sys.path.append("/workspaces/core/charles_dev")
import json
from logger import logger
import os
import socketio
from daikin_api import DaikinAPI
from daikin_installation import DaikinInstallation
from daikin_device import DaikinDevice
from daikin_profile import DaikinProfile


class DaikinCloud:
    """Methods to interact with the Daikin One API."""

    data_file = "tmp/daikin.json"
    api: DaikinAPI
    installations: dict[str, DaikinInstallation]
    devices: list[DaikinDevice]
    user_socket: socketio.AsyncClient
    profile: DaikinProfile

    def __init__(self) -> None:
        self.api = DaikinAPI()
        self.user_socket = socketio.AsyncClient()
        self.devices = []

    async def after_login(self):
        await self.fetch_installations()
        await self.connect_user_socket()

    async def login_dev(self):
        self.load_environment()
        email = os.environ.get("daikin_email")
        password = os.environ.get("daikin_password")
        response = await self.api.login_pasword(email, password)
        self.profile = response["profile"]
        self.save_environment()
        await self.after_login()

    async def login(self, email: str, password: str):
        response = await self.api.login_pasword(email, password)
        self.profile = response["profile"]
        await self.after_login()

    def load_environment(self):
        """Loads tokens from OS environment variables; useful for REPL."""
        try:
            with open(self.data_file, encoding="UTF-8") as read_file:
                data = json.load(read_file)
                self.api.set_tokens(data["tokens"])
        except (KeyError, FileNotFoundError, json.decoder.JSONDecodeError):
            self.token = ""
            self.refresh_token = ""
            logger.debug("No tokens to load from environment")

    def save_environment(self):
        """Saves tokens from OS environment variables; useful for REPL."""
        with open(self.data_file, "w", encoding="UTF-8") as write_file:
            json.dump({"tokens": self.api.get_tokens()}, write_file)
        logger.debug("Saved tokens to environment")

    async def fetch_installations(self):
        """Gets installations available to the current user."""
        logger.debug("Getting installations")
        response = await self.api.get(f"installations/{self.api.SCOPE}")
        self.installations: dict[str, DaikinInstallation] = {}
        for i in response:
            thisInstallation = DaikinInstallation(self.api, i)
            await thisInstallation.connect_installation_socket()
            self.installations[i["_id"]] = thisInstallation
            self.devices.extend(thisInstallation.devices.values())

    async def connect_user_socket(self):
        """connects the user socket"""

        @self.user_socket.event(namespace="/users")
        def message(data):
            logger.debug("User socket event: '%s'", json.dumps(data))

        @self.user_socket.on("*")
        def catch_all(event, data):
            """Default event handler"""
            logger.debug(
                "User socket message '%s': %s",
                json.dumps(event),
                json.dumps(data),
            )

        @self.user_socket.event(namespace="/users")
        def connect():
            """SIO Connect callback"""
            logger.debug("User socket connected!")

        @self.user_socket.event(namespace="/users")
        def connect_error(data):
            """SIO Connect Error callback"""
            logger.debug("User socket connection failed!")

        @self.user_socket.event(namespace="/users")
        def disconnect():
            """SIO Disonnect callback"""
            logger.debug("User socket disconnected!")

        url = f"{self.api.API_URL}"
        logger.debug("Starting user socket: %s", url)
        await self.user_socket.connect(
            url,
            namespaces=["/users"],
            transports=["polling"],
            socketio_path=self.api.SOCKET_PATH,
            headers={"Authorization": f"Bearer {self.api.api_tokens['access_token']}"},
        )



"""Module: """
from __future__ import annotations
from logger import logger
from typing import TYPE_CHECKING
from dataclasses import fields

from collections.abc import Callable
import json
from daikin_device_data import DeviceData

if TYPE_CHECKING:
    from daikin_installation import DaikinInstallation


class DaikinDevice:
    """Class for interacting with device"""

    mac: str
    icon: str
    name: str
    acs_mode: bool
    installation: DaikinInstallation
    device_data: DeviceData
    on_data_updated: list[Callable[[]]]

    def __init__(self, device_data: any, installation: DaikinInstallation) -> None:
        self.acs_mode = False
        self.name = device_data["name"]
        self.icon = device_data["icon"]
        self.mac = device_data["mac"]
        self.installation = installation
        self.device_data = DeviceData()
        self.on_data_updated = []
        logger.debug("Setup device %s - %s", self.name, self.mac)

    async def set_device_value(self, prop: str, value: str):
        """Sets a device value using the installation socket"""

        update_command = {"mac": self.mac, "property": prop, "value": value}

        def callback_fun(data):
            logger.debug(
                "(%s) POST - %s : %s; (raw: %s)",
                update_command["mac"],
                update_command["property"],
                update_command["value"],
                json.dumps(data),
            )

        logger.debug("Setting device '%s' value '%s' to '%s'", self.mac, prop, value)

        await self.installation.emit(
            "create-machine-event", update_command, callback=callback_fun
        )
        setattr(self.device_data, prop, value)

    def add_update_callback(self, callback: Callable[[]]):
        """Updates the callback and then calls it"""
        self.on_data_updated.append(callback)
        self._invoke_callback()

    def handle_device_data(self, data: DeviceData):
        """Update this device model with new data from Daikin Cloud"""
        logger.debug("Got device Data for '%s'", self.name)
        changes = False
        for field in fields(data):
            new_value = getattr(data, field.name)
            if new_value is not None:
                setattr(self.device_data, field.name, new_value)
                logger.debug(
                    "Device property '%s' changed to '%s'", field.name, new_value
                )
                changes = True
        if changes:
            if hasattr(self, "on_data_updated") and self.on_data_updated is not None:
                self._invoke_callback()

    def _invoke_callback(self):
        if self.on_data_updated.__len__ == 0:
            logger.debug("No callbacks registered")
            return

        logger.debug("Notifying al on_data_updated callbacks of changes")
        for callback in self.on_data_updated:
            try:
                callback()
            except Exception:
                logger.exception("Failed to execute a callback")



"""test program"""
import asyncio
from dotenv import load_dotenv
from daikin_cloud import DaikinCloud
from logger import logger, log_to_console
import json


loop = asyncio.get_event_loop()


log_to_console()


async def main():
    """runner"""
    logger.debug("Loading environment")
    load_dotenv()
    logger.debug("Initializing DaikinCloud")
    d = DaikinCloud()
    await d.login_dev()
    logger.debug("profile %s", d.profile.email)


loop.create_task(main())
loop.run_forever()



"""Module: """
import json
from logger import logger
import socketio
from daikin_device import DaikinDevice
from daikin_api import DaikinAPI
from daikin_device_data import DeviceDataMessage


class DaikinInstallation:
    """Class for interacting with an installation"""

    installation_socket: socketio.AsyncClient
    api: DaikinAPI
    installation_id: str
    devices: dict[str, DaikinDevice]
    installation_namespace: str

    def __init__(self, api, installation_data) -> None:
        """Sets up this installation"""
        self.api = api
        self.devices = []
        self.installation_id = installation_data["_id"]
        self.devices: dict[str, DaikinDevice] = {}
        self.installation_namespace = f"/{self.installation_id}::{self.api.SCOPE}"
        self.installation_socket = socketio.AsyncClient()
        for d in installation_data["devices"]:
            self.devices[d["mac"]] = DaikinDevice(d, self)
            logger.debug(
                "Set up device '%s':'%s' in installation '%s'",
                d["name"],
                d["mac"],
                self.installation_id,
            )
        logger.debug("Set up installation '%s'", self.installation_id)

    async def emit(self, event, data, callback):
        """Proxies the Socket.IO emit but includes this installation's namespace"""
        await self.installation_socket.emit(
            event, data, namespace=self.installation_namespace, callback=callback
        )

    async def connect_installation_socket(self):
        """starts the Socket.IO connection for the provided installation"""

        url = f"{self.api.API_URL}"
        # url = "http://172.30.9.155:3000"

        logger.debug(
            "Starting installation socket: %s; namespace: '%s'",
            url,
            self.installation_namespace,
        )
        await self.installation_socket.connect(
            url,
            transports=["polling"],
            namespaces=[self.installation_namespace],
            socketio_path=self.api.SOCKET_PATH,
            headers={"Authorization": f"Bearer {self.api.api_tokens['access_token']}"},
        )

        @self.installation_socket.on("*", namespace=self.installation_namespace)
        def catch_all(event, data):
            """Default event handler"""
            logger.debug(
                "Installation socket message '%s': %s",
                json.dumps(event),
                json.dumps(data),
            )

        @self.installation_socket.on(
            "device-data", namespace=self.installation_namespace
        )
        def on_device_data(data):
            device_data_message: DeviceDataMessage = DeviceDataMessage.from_dict(data)
            try:
                self.devices[device_data_message.mac].handle_device_data(
                    device_data_message.data
                )
            except KeyError:
                logger.warning(
                    "Device '%s' does not exist in installation '%s'",
                    device_data_message.mac,
                    self.installation_id,
                )

        @self.installation_socket.event(namespace=self.installation_namespace)
        def message(data):
            logger.debug("Installation socket message '%s'", data)

        @self.installation_socket.event
        def connect():
            """installation_socket Connect callback"""
            logger.debug(
                "Installation socket connected: '%s'",
                self.installation_namespace,
            )

        @self.installation_socket.event
        def connect_error(data):
            """installation_socket Connect Error callback"""
            logger.debug("Installation socket connection failed!")

        @self.installation_socket.event
        def disconnect():
            """installation_socket Disonnect callback"""
            logger.debug("Installation socketdisconnected!")



"""Module"""
from dataclasses import dataclass
from dataclasses_json import dataclass_json
from typing import Optional


@dataclass_json
@dataclass
class DeviceManufacturer:
    _id: Optional[int] = None
    text: Optional[str] = None


@dataclass_json
@dataclass
class DeviceData:
    machineready: Optional[bool] = None
    version: Optional[str] = None
    aidooit: Optional[bool] = None
    emerheatpresent: Optional[bool] = None
    emerheatstatus: Optional[bool] = None
    fallback: Optional[bool] = None
    t1t2on: Optional[bool] = None
    real_mode: Optional[int] = None
    work_temp_selec_sensor: Optional[int] = None
    stat_channel: Optional[int] = None
    stat_rssi: Optional[int] = None
    stat_ssid: Optional[str] = None
    manufacturer: Optional[DeviceManufacturer] = None
    power: Optional[bool] = None
    mode: Optional[int] = None
    mode_available: Optional[any] = None
    speed_available: Optional[any] = None
    speed_state: Optional[int] = None
    slats_autoud: Optional[bool] = None
    slats_swingud: Optional[bool] = None
    slats_vnum: Optional[int] = None
    range_sp_cool_air_max: Optional[int] = None
    range_sp_cool_air_min: Optional[int] = None
    range_sp_hot_air_max: Optional[int] = None
    range_sp_hot_air_min: Optional[int] = None
    range_sp_auto_air_max: Optional[int] = None
    range_sp_auto_air_min: Optional[int] = None
    device_master_slave: Optional[bool] = None
    master: Optional[bool] = None
    setpoint_step: Optional[bool] = None
    units: Optional[int] = None
    setpoint_air_cool: Optional[int] = None
    setpoint_air_heat: Optional[int] = None
    setpoint_air_auto: Optional[int] = None
    error_value: Optional[int] = None
    error_ascii1: Optional[str] = None
    error_ascii2: Optional[str] = None
    tsensor_error: Optional[bool] = None
    work_temp: Optional[int] = None
    icon: Optional[int] = None
    name: Optional[str] = None
    timezoneId: Optional[str] = None
    isConnected: Optional[bool] = None


@dataclass_json
@dataclass
class DeviceDataMessage:
    mac: Optional[str] = None
    data: Optional[DeviceData] = None



"""Logging module"""
import logging

logger = logging.getLogger("Daikin")
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.NOTSET)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
ch.setFormatter(formatter)

log_everything = False


def log_to_console():
    logger.addHandler(ch)
    logger.debug("Logger Initialized")


if log_everything:
    logging.basicConfig(level=logging.NOTSET)



"""Daikin API class"""
import asyncio

from daikin_profile import DaikinProfile
from logger import logger
import requests
from typing_extensions import TypedDict
import functools

APITokens = TypedDict("APITokens", {"access_token": str, "refresh_token": str})

# TODO:self.authHeader = headers["Authorization"] in charles_dev/tmp/site-packages/engineio/client.py


class DaikinAPI:
    """Commands for interacting with REST and Socket.IO"""

    api_tokens: APITokens
    timeout_seconds = 5
    API_URL = "https://dkncloudna.com/"
    API_VERSION = "api/v1/"
    SCOPE = "dknUsa"
    SOCKET_PATH = "/api/v1/devices/socket.io/"
    loop: asyncio.AbstractEventLoop

    def __init__(self) -> None:
        self.api_tokens = {}
        self.loop = asyncio.get_event_loop()

    def build_headers(self, headers: dict = None, auth: bool = True):
        """Creates headers for API requests"""
        if headers is None:
            headers = {}
        headers["Access-Control-Allow-Origin"] = "*"
        if auth:
            headers["Authorization"] = f"Bearer {self.api_tokens['access_token']}"
        return headers

    async def post(
        self, subpath: str, headers: dict = None, body: dict = None, auth: bool = True
    ):
        """Proxy api post method."""
        headers = self.build_headers(headers, auth)
        url = f"{self.API_URL}{self.API_VERSION}{subpath}"
        logger.debug("API POST to %s", url)
        request = await self.loop.run_in_executor(
            None,
            functools.partial(
                requests.post,
                url,
                headers=headers,
                json=body,
                timeout=self.timeout_seconds,
            ),
        )
        return request.json()

    async def get(self, subpath: str, headers: dict = None, auth: bool = True):
        """Proxy api get method."""
        headers = self.build_headers(headers, auth)
        url = f"{self.API_URL}{self.API_VERSION}{subpath}"
        logger.debug("API GET to %s", url)
        request = await self.loop.run_in_executor(
            None,
            functools.partial(
                requests.get, url, headers=headers, timeout=self.timeout_seconds
            ),
        )
        return request.json()

    async def login_pasword(self, email: str, password: str):
        """Logs in and gets a token / refreshtoken."""
        # if "access_token" in self.api_tokens and self.api_tokens["access_token"]:
        # logger.debug("Already logged in")
        # return

        if email is None:
            raise ValueError("No email address")
        if password is None:
            raise ValueError("No password")

        login_path = f"auth/login/{self.SCOPE}"
        logger.debug("Logging in as %s", email)
        response = await self.post(
            login_path, body={"email": email, "password": password}, auth=False
        )
        self.api_tokens = {
            "access_token": response["token"],
            "refresh_token": response["refreshToken"],
        }
        logger.debug("Logged in as %s", email)
        return {
            "profile": DaikinProfile.from_dict(response["data"]),
            "installations": response["scope"][self.SCOPE]["installations"],
        }

    def set_tokens(self, api_tokens: APITokens):
        """Sets api tokens"""
        self.api_tokens = api_tokens

    def get_tokens(self) -> APITokens:
        """gets api tokens"""
        return self.api_tokens

    def update_token(self):
        """Updates the current token using the refresh token."""
        response = self.get(
            f"auth/refreshToken/{self.api_tokens['refresh_token']}/{self.SCOPE}"
        )
        logger.debug(
            "Updating access token using refresh token %s",
            self.api_tokens["refresh_token"],
        )
        self.api_tokens["access_token"] = response["token"]
        self.api_tokens["refresh_token"] = response["refreshToken"]
        logger.debug("Token refresh success")
